# -*- coding: utf-8 -*-
import json

import requests
from flask import current_app
from pydantic.main import ModelMetaclass, BaseModel
from requests.adapters import HTTPAdapter

from easi_py_common.core.error import ServiceException

HEADER = {'Content-Type': 'application/json; charset=UTF-8'}
DST_MODE_SERVICE = 'service'
DST_MODE_OTHER = 'other'


class RestClient:
    def __init__(self, url, timeout=(5, 5), http_adapter=None, before_request=None,
                 dst_mode=DST_MODE_SERVICE):
        self.url = url
        if http_adapter is None:
            http_adapter = HTTPAdapter(pool_connections=10, pool_maxsize=10, max_retries=3)
        session = requests.Session()
        session.mount('http://', http_adapter)
        session.mount('https://', http_adapter)

        self.session = session
        self.timeout = timeout
        self.before_request = before_request
        self.dst_mode = dst_mode

    def _get_url(self, uri):
        if 'SERVICE_CLIENT' in current_app.config and 'http' not in self.url:
            service_client_keys = current_app.config['SERVICE_CLIENT']
            self.url = self.url.format(**service_client_keys)
        return self.url + uri

    def get(self, uri, params=None, timeout=None, dst=None, **kwargs):
        if timeout is None:
            timeout = self.timeout

        _url = self._get_url(uri)

        if callable(self.before_request):
            _request_parms = {
                'url': _url,
                'method': 'GET',
                'params': params,
            }
            _kwargs = self.before_request(_request_parms)
            kwargs.update(_kwargs)

        resp = self.session.get(
            _url,
            params=params,
            timeout=timeout,
            **kwargs
        )
        return self._do_response(resp, dst=dst)

    def post(self, uri, json_data=None, params=None, timeout=None, dst=None, **kwargs):
        if timeout is None:
            timeout = self.timeout

        _url = self._get_url(uri)
        data = None
        if json_data:
            data = (json_data.json(exclude_none=True, by_alias=False)
                    if isinstance(json_data, BaseModel)
                    else json.dumps(json_data))

        if callable(self.before_request):
            _request_parms = {
                'url': _url,
                'method': 'POST',
                'data': data,
            }
            _kwargs = self.before_request(_request_parms)
            kwargs.update(_kwargs)

        if 'headers' in kwargs:
            kwargs.get('headers').update(HEADER)
        else:
            kwargs['headers'] = HEADER

        resp = self.session.post(
            _url,
            data=data,
            params=params,
            timeout=timeout,
            **kwargs
        )
        return self._do_response(resp, dst=dst)

    def put(self, uri, json_data=None, params=None, timeout=None, dst=None, **kwargs):
        if timeout is None:
            timeout = self.timeout

        _url = self._get_url(uri)

        data = None
        if json_data:
            data = (json_data.json(exclude_none=True, by_alias=False)
                    if isinstance(json_data, BaseModel)
                    else json.dumps(json_data))

        if callable(self.before_request):
            _request_parms = {
                'url': _url,
                'method': 'PUT',
                'data': data,
            }
            _kwargs = self.before_request(_request_parms)
            kwargs.update(_kwargs)

        if 'headers' in kwargs:
            kwargs.get('headers').update(HEADER)
        else:
            kwargs['headers'] = HEADER

        resp = self.session.put(
            _url,
            data=data,
            params=params,
            timeout=timeout,
            **kwargs
        )
        return self._do_response(resp, dst=dst)

    def delete(self, uri, params=None, timeout=None, dst=None, **kwargs):
        if timeout is None:
            timeout = self.timeout

        _url = self._get_url(uri)
        if callable(self.before_request):
            _request_parms = {
                'url': _url,
                'method': 'DELETE',
            }
            _kwargs = self.before_request(_request_parms)
            kwargs.update(_kwargs)

        resp = self.session.delete(
            _url,
            params=params,
            timeout=timeout,
            **kwargs
        )
        return self._do_response(resp, dst=dst)

    def _do_response(self, response, dst=None):
        status_code = response.status_code
        if status_code == 200:
            result = response.json()
            if (
                    isinstance(result, bool)
                    or isinstance(result, int)
                    or isinstance(result, float)
                    or isinstance(result, str)
            ):
                return result

            if self.dst_mode == DST_MODE_SERVICE:
                code = result.get('code', 0)
                if code != 0:
                    raise ServiceException(code=code, msg=result.get('message', ''))
                result = result.get('data')

            if dst and dst.__class__ == ModelMetaclass:
                return dst.parse_obj(result)
            return result

        if 400 <= status_code < 500:
            raise ServiceException(code=status_code, msg=response.content)
        else:
            raise ServiceException(code=status_code, msg=response.content)



import json
import time
import uuid

import boto3
import orjson
from botocore.config import Config

DEAD_QUEUE_NAME = 'easipay_letter_dead'
DEAD_QUEUE_NAME_FIFO = 'easipay_letter_dead.fifo'


def get_utc_timestamp():
    timestamp = time.time()
    return int(timestamp * 1000)


class SQSClient:
    __queue_map = {}
    __resource = None
    __dead_queue = None
    __dead_queue_fifo = None

    def __init__(self, region_name,
                 queue_name_prefix='',
                 aws_access_key_id='',
                 aws_secret_access_key='',
                 max_pool_connections=50,
                 max_attempts=3):
        self.queue_name_prefix = queue_name_prefix
        config = Config(
            region_name=region_name,
            signature_version='v4',
            max_pool_connections=max_pool_connections,
            retries={
                'max_attempts': max_attempts,
                # 'mode': 'standard'
            }
        )

        self.__client = boto3.client(service_name='sqs',
                                     aws_access_key_id=aws_access_key_id,
                                     aws_secret_access_key=aws_secret_access_key,
                                     config=config)
        self.__init_queue()

    def send_message(self, queue_name, body, key='', group_id='', delay_seconds=0, cb=None):
        if '.fifo' in queue_name:
            if not group_id:
                raise Exception('MessageGroupId is required')

        return self.__send_message(queue_name, body, key, group_id, delay_seconds, cb, get_utc_timestamp())

    def set_queue_attributes(self, queue_url, attributes):
        self.__client.set_queue_attributes(
            QueueUrl=queue_url,
            Attributes=attributes)

    def get_queue_url(self, queue_name):
        url = self.__get_url_by_name(queue_name)
        return url

    def get_queue_attr(self, qurl, attr_names):
        return self.__client.get_queue_attributes(
            QueueUrl=qurl,
            AttributeNames=attr_names
        )

    def get_sqs_client(self):
        return self.__client

    def __init_queue(self):
        if len(self.queue_name_prefix) > 0:
            queue_data = self.__client.list_queues(QueueNamePrefix=self.queue_name_prefix)
        else:
            queue_data = self.__client.list_queues()

        if 'QueueUrls' in queue_data:
            for queue_url in queue_data['QueueUrls']:
                name = self.__get_name_by_url(queue_url)
                self.__queue_map[name] = queue_url

        if not self.__dead_queue:
            self.__install_dead(DEAD_QUEUE_NAME)
            self.__install_dead(DEAD_QUEUE_NAME_FIFO)

    def __get_name_by_url(self, q_url):
        return q_url[q_url.rindex('/') + 1:]

    def __get_url_by_name(self, queue_name):
        if queue_name not in self.__queue_map:
            qurl = self.__get_queue_url(queue_name)
            self.__queue_map[queue_name] = qurl
        return self.__queue_map[queue_name]

    def __send_message(self, queue_name, body, key='', group_id='', delay_seconds=0, cb=None,
                       send_start_time=get_utc_timestamp()):
        execute_start_time = get_utc_timestamp()
        work_wait_time = int(execute_start_time - send_start_time)
        msg_body = dict(
            t=execute_start_time,
            msg=body,
            key=key
        )
        try:
            queue_url = self.__get_queue_url(queue_name)

            if '.fifo' in queue_name:
                response = self.__client.send_message(
                    QueueUrl=queue_url,
                    MessageBody=json.dumps(msg_body),
                    MessageGroupId=str(group_id),
                    MessageDeduplicationId=str(uuid.uuid1())
                )
            else:
                if delay_seconds > 0:
                    if delay_seconds > 7200:
                        delay_seconds = 7200
                    msg_body['delay_seconds'] = delay_seconds

                if delay_seconds > 900:
                    delay_seconds = 900

                response = self.__client.send_message(
                    QueueUrl=queue_url,
                    DelaySeconds=delay_seconds,
                    MessageBody=json.dumps(msg_body),
                )

            msg_id = ''
            if 'MessageId' in response:
                msg_id = response['MessageId']

            if 'Code' in response:
                msg_id = response['Code']

            if cb:
                log_data = dict(
                    msg=msg_body['msg'],
                    t=msg_body['t'],
                    s_t=get_utc_timestamp(),
                    key=key,
                    group_id=group_id,
                    url=queue_url,
                    address=queue_name,
                    delay_seconds=msg_body['delay_seconds'] if 'delay_seconds' in msg_body else 0,
                    work_wait_time=work_wait_time
                )
                cb(msg_id, log_data)
                # self.__pool_work_log.submit(cb, msg_id, log_data)
            return True
        except Exception as e:
            if cb:
                log_data = dict(
                    msg=msg_body['msg'],
                    t=msg_body['t'],
                    s_t=get_utc_timestamp(),
                    key=key,
                    group_id=group_id,
                    url=e.message[0:254],
                    address=queue_name,
                    delay_seconds=msg_body['delay_seconds'] if 'delay_seconds' in msg_body else 0,
                    work_wait_time=work_wait_time,
                )

                cb('fail', log_data)
        return False

    def send_message_batch(self, queue_name, msgs, cb=None):
        if queue_name not in self.__queue_map:
            qurl = self.__get_queue_url(queue_name)
            self.__queue_map[queue_name] = qurl

        queue_url = self.__queue_map[queue_name]
        is_fifo = False
        if '.fifo' in queue_name:
            is_fifo = True

        entries = []
        for msg in msgs:
            if is_fifo:
                entries.append({
                    'MessageBody': orjson.dumps(msg['body']),
                    'MessageGroupId': str(msg['group_id']),
                    'MessageDeduplicationId': str(uuid.uuid1()),
                })
            else:
                delay_seconds = msg.get('delay_seconds', 0)
                if delay_seconds < 0:
                    delay_seconds = 0

                if delay_seconds > 0:
                    if delay_seconds > 7200:
                        delay_seconds = 7200

                if delay_seconds > 900:
                    delay_seconds = 900

                entries.append({
                    'MessageBody': orjson.dumps(msg['body']),
                    'DelaySeconds': delay_seconds

                })
        response = None
        try:
            response = self.__client.send_message_batch(QueueUrl=queue_url, Entries=entries)
        except Exception as e:
            if cb:
                cb('fail', {
                    'response': response,
                    'queue_url': queue_url,
                    'entries': entries
                })

    def __get_queue_url(self, queue_name, MessageRetentionPeriod=86400 * 3, receive_message_wait_time_seconds=10,
                        visibility_timeout=300, max_receive_count=6):
        if queue_name in self.__queue_map:
            return self.__queue_map[queue_name]

        try:
            queue = self.__client.get_queue_url(
                QueueName=queue_name,
            )
            if 'QueueUrl' in queue:
                return queue['QueueUrl']
        except:
            pass

        attributes = {
            'VisibilityTimeout': str(visibility_timeout),
            'DelaySeconds': '0',
            'MessageRetentionPeriod': str(MessageRetentionPeriod),
            'ReceiveMessageWaitTimeSeconds': str(receive_message_wait_time_seconds)
        }

        if '.fifo' in queue_name:
            attributes['FifoQueue'] = 'true'
            if self.__dead_queue_fifo:
                attributes['RedrivePolicy'] = json.dumps({
                    'deadLetterTargetArn': self.__dead_queue_fifo['arn'],
                    'maxReceiveCount': str(max_receive_count)
                })
        else:
            if self.__dead_queue:
                attributes['RedrivePolicy'] = json.dumps({
                    'deadLetterTargetArn': self.__dead_queue['arn'],
                    'maxReceiveCount': str(max_receive_count)
                })

        queue = self.__client.create_queue(
            QueueName=queue_name,
            Attributes=attributes
        )
        queue_url = queue['QueueUrl']
        return queue_url

    def __install_dead(self, dead_name):
        dead_queue_url = self.__get_queue_url(dead_name, MessageRetentionPeriod=86400 * 10)

        response = self.__client.get_queue_attributes(
            QueueUrl=dead_queue_url,
            AttributeNames=['QueueArn']
        )
        if '.fifo' in dead_name:
            self.__dead_queue_fifo = {'url': dead_queue_url, 'arn': response['Attributes']['QueueArn']}
        else:
            self.__dead_queue = {'url': dead_queue_url, 'arn': response['Attributes']['QueueArn']}



import json

import orjson
import traceback

import flask
from gevent.pool import Pool
from datetime import datetime

from easi_py_common.core.redis import redis_store
from easi_py_common.sqs import constants
from easi_py_common.sqs.constants import ConsumerLogConstants
from easi_py_common.sqs.producer_session import ProducerSession
from easi_py_common.sqs.sqs import SQSClient

pool = Pool(128)

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


class Producer:

    def __init__(self):
        self.started = False
        self.sqs_client = None
        self.app = None
        self.sign = ''

    def __format_msg(self, msg):
        if isinstance(msg, (str, int, bool, float)):
            return msg
        else:
            return orjson.dumps(msg)

    def start(self, app: flask.app.Flask, max_pool_connections=50, max_attempts=3):
        self.app = app

        if self.started:
            return

        self.started = True

        if self.app and self.app.config['ENV'] != 'prod':
            if 'sign' in app.config['PUBSUB']:
                self.sign = app.config['PUBSUB']['sign'] + '_'

        if 'region_name' in app.config['PUBSUB']:
            try:
                aws_access_key_id = app.config['AWS_ACCESS_KEY_ID']
                aws_secret_access_key = app.config['AWS_SECRET_ACCESS_KEY']
                self.sqs_client = SQSClient(app.config['PUBSUB']['region_name'],
                                            queue_name_prefix=self.sign,
                                            aws_access_key_id=aws_access_key_id,
                                            aws_secret_access_key=aws_secret_access_key,
                                            max_pool_connections=max_pool_connections,
                                            max_attempts=max_attempts)
            except Exception:
                if self.app:
                    self.app.logger.error(traceback.format_exc())

        self.init_batch_mode(self.app)

    def start_by_sqs_client(self, sqs_client):
        self.sqs_client = sqs_client

    def publish(self, address, msg, key='', group_id='', delay_seconds=0, sign=None):
        pool.spawn(self.send, address, msg, key=key, group_id=group_id, delay_seconds=delay_seconds, sign=sign)

    def send(self, address, msg, key='', group_id='', delay_seconds=0, sign=None):
        if self.sqs_client is None:
            if self.app:
                self.app.logger.error("sqs_client is None")
            return False
        if sign is None:
            address = self.sign + address
        else:
            address = sign + address
        return self.sqs_client.send_message(address, msg, key=key,
                                            group_id=group_id, delay_seconds=delay_seconds, cb=self.__log)

    def get_sqs_client(self):
        return self.sqs_client

    @property
    def session(self):
        # type: () -> ProducerSession
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'producer_session'):
                ctx.producer_session = ProducerSession(self.sqs_client, self.sign, app=self.app, cb=None)
            return ctx.producer_session

    def commit(self):
        self.session.commit()

    def init_batch_mode(self, app):
        if hasattr(app, 'teardown_appcontext'):
            app.teardown_appcontext(self.batch_mode_teardown)
        else:
            app.teardown_request(self.batch_mode_teardown)

    def batch_mode_teardown(self, exception):
        ctx = stack.top
        if hasattr(ctx, 'producer_session'):
            ctx.producer_session.commit()
            # if exception is None:
            #     ctx.producer_session.commit()
            # else:
            #     ctx.producer_session.flush()
        return exception

    def __log(self, msg_id, msg_body):
        send_time = self.__format_timestamp(msg_body['t'])
        send_success_time = self.__format_timestamp(msg_body['s_t'])
        msg_body['send_time'] = send_time
        msg_body['send_success_time'] = send_success_time
        msg_body['msg_id'] = msg_id

        producer_log_len = redis_store.lpush(ConsumerLogConstants.PUB_LOG, json.dumps(msg_body))
        if producer_log_len >= ConsumerLogConstants.CONSUMER_LOG_MAX_SLEEP_ES_BULK_BUFFER_SIZE:
            redis_store.rpop(ConsumerLogConstants.PUB_LOG)
            return
        # 计算是否可以进行保存
        can_bulk_save = producer_log_len % ConsumerLogConstants.CONSUMER_LOG_ES_BULK_NUM == 0
        # 如果不可以保存，并且小于缓存最大值，则直接返回
        if not can_bulk_save and producer_log_len < ConsumerLogConstants.CONSUMER_LOG_ES_BULK_BUFFER_SIZE:
            return
        # 如果可以保存 或者 是200的倍数 则尝试进行保存
        if can_bulk_save or producer_log_len % 200 == 0 or producer_log_len >= ConsumerLogConstants.CONSUMER_LOG_MAX_SLEEP_ES_BULK_BUFFER_SIZE:
            self.publish(constants.PUB_LOG_BULK_SAVE, '')
            return

    def __format_timestamp(self, timestamp, is_millisecond=True):
        if is_millisecond:
            timestamp = round(timestamp / 1000.0, 2)
        date_array = datetime.utcfromtimestamp(timestamp)
        return date_array.strftime('%Y.%m.%d %H:%M:%S.%f')[:-3]


producer = Producer()



import json
import sys
import traceback

from flask import jsonify, make_response, current_app
from functools import wraps
from pydantic import ValidationError
from sqlalchemy.exc import IntegrityError


class ServiceException(Exception):
    """服务异常
    主要针对于一些逻辑异常，就抛出相关的错误。
    比如当前用户不存在，名称不能为空
    """

    def __init__(self, code=400, msg='', data=None, *args, **kwargs):
        self.code = code
        self.message = msg
        self.data = data
        super(ServiceException, self).__init__(*args, **kwargs)

    def __repr__(self):
        return u"{}<code:{}, message:{}>".format(self.__class__.__name__, self.code, self.message)

    def __unicode__(self):
        return u"{}<code:{}, message:{}>".format(self.__class__.__name__, self.code, self.message)

class InvalidParmsException(ServiceException):
    def __init__(self):
        super(InvalidParmsException, self).__init__(code=20000, msg='Invalid parameter')

class DataNotFoundException(ServiceException):

    def __init__(self):
        super(DataNotFoundException, self).__init__(msg='data is not found')


class InvalidTokenException(ServiceException):
    def __init__(self):
        super(InvalidTokenException, self).__init__(code=20001, msg='invalid access token')


class TokenExpiredException(ServiceException):
    def __init__(self):
        super(TokenExpiredException, self).__init__(code=20000, msg='token expired')


def intercept_service_exception(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            ret = f(*args, **kwargs)
            return ret
        except ServiceException as e:
            if e.data:
                return jsonify(code=e.code, message=e.message, data=e.data)
            return jsonify(code=e.code, message=e.message)

        except ValidationError as v:

            try:
                traceback_exception_list = traceback.format_exception(*sys.exc_info())
                if traceback_exception_list and len(traceback_exception_list) >= 3 and traceback_exception_list[-3]:
                    current_app.logger.error('validation error: {}\n{}'.format(
                        json.dumps(v.errors()), traceback_exception_list[-3].split('\n')[0]))
            except Exception:
                pass

            return make_response(jsonify({'code': 100, 'message': 'validation_error',
                                          "validation error": v.errors()}), 200)
        except IntegrityError:
            return jsonify(code=1100, message=u'重复操作，请稍后再试')

    return wrapped

from typing import List


class BaseFlaskPydanticException(Exception):
    """Base exc class for all exception from this library"""

    pass


class InvalidIterableOfModelsException(BaseFlaskPydanticException):
    """This exception is raised if there is a failure during serialization of
    response object with `response_many=True`"""

    pass


class JsonBodyParsingError(BaseFlaskPydanticException):
    """Exception for error occurring during parsing of request body"""

    pass


class ManyModelValidationError(BaseFlaskPydanticException):
    """This exception is raised if there is a failure during validation of many
    models in an iterable"""

    def __init__(self, errors: List[dict], *args):
        self._errors = errors
        super().__init__(*args)

    def errors(self):
        return self._errors
